from .request import Request
from .response import Response
from .stream import Stream

class Router:
    __slots__ = ('_routes',)

    def __init__(self):
        self._routes = {}

    def get(self, path):
        def decorator(fn):
            self._routes[("GET", path)] = fn
            return fn
        return decorator

    def post(self, path):
        def decorator(fn):
            self._routes[("POST", path)] = fn
            return fn
        return decorator

    async def __call__(self, scope, receive, send):
        if scope["type"] == "lifespan":
            while True:
                msg = await receive()
                if msg["type"] == "lifespan.startup":
                    await send({"type": "lifespan.startup.complete"})
                elif msg["type"] == "lifespan.shutdown":
                    await send({"type": "lifespan.shutdown.complete"})
                    return
            return
        if scope["type"] != "http": return
        req = Request(scope, receive)
        key = (req.method, req.path)
        handler = self._routes.get(key)
        if not handler:
            res = Response(send)
            await res.text("Not Found", status=404)
            return
        await handler(req, send)



"""Minimal Async implementation for sse"""
__version__ = '0.0.2'
__author__ = 'Deufel'
from .request import Request
from .response import Response
from .stream import Stream
from .relay import Relay
from .relay import Router
__all__ = [
    "Relay",
    "Request",
    "Response",
    "Router",
    "Stream",
]



class Response:
    __slots__ = ('_send', '_headers')

    def __init__(self, send):
        self._send = send
        self._headers = []

    def cookie(self, name, value, **opts):
        parts = [f"{name}={value}"]
        for k, v in opts.items():
            k = k.replace("_", "-")
            if isinstance(v, bool):
                if v: parts.append(k)
            else: parts.append(f"{k}={v}")
        self._headers.append([b"set-cookie", "; ".join(parts).encode()])
        return self

    async def _send_response(self, status, content_type, body):
        headers = [[b"content-type", content_type]] + self._headers
        await self._send({"type": "http.response.start", "status": status, "headers": headers})
        await self._send({"type": "http.response.body", "body": body.encode() if isinstance(body, str) else body})

    async def html(self, body, status=200):
        await self._send_response(status, b"text/html; charset=utf-8", body)

    async def text(self, body, status=200):
        await self._send_response(status, b"text/plain; charset=utf-8", body)

    async def redirect(self, url, status=302):
        headers = [[b"location", url.encode()]] + self._headers
        await self._send({"type": "http.response.start", "status": status, "headers": headers})
        await self._send({"type": "http.response.body", "body": b""})



import asyncio

class Stream:
    __slots__ = ('_send', '_headers', '_closed')

    def __init__(self, send):
        self._send = send
        self._headers = []
        self._closed = False

    def cookie(self, name, value, **opts):
        parts = [f"{name}={value}"]
        for k, v in opts.items():
            k = k.replace("_", "-")
            if isinstance(v, bool):
                if v: parts.append(k)
            else: parts.append(f"{k}={v}")
        self._headers.append([b"set-cookie", "; ".join(parts).encode()])
        return self

    async def open(self):
        headers = [
            [b"content-type", b"text/event-stream"],
            [b"cache-control", b"no-cache"],
            [b"connection", b"keep-alive"],
            [b"x-accel-buffering", b"no"],
        ] + self._headers
        await self._send({"type": "http.response.start", "status": 200, "headers": headers})

    async def send_event(self, data):
        "Send a pre-formatted SSE event string (from html-tags sse module)"
        if self._closed: return
        try:
            await self._send({"type": "http.response.body", "body": data.encode(), "more_body": True})
        except Exception:
            self._closed = True

    async def keepalive(self):
        "Send an SSE comment to keep the connection alive"
        await self.send_event(":\n\n")

    async def close(self):
        if self._closed: return
        self._closed = True
        await self._send({"type": "http.response.body", "body": b""})

    async def alive(self, events, timeout=30):
        "Yield relay events while keeping the connection alive"
        async for event in events:
            if self._closed: break
            yield event



from urllib.parse import parse_qs

class Request:
    __slots__ = ('path', 'method', 'headers', 'query', 'cookies', '_receive')

    def __init__(self, scope, receive):
        self.path = scope["path"]
        self.method = scope.get("method", "GET")
        self._receive = receive
        # Parse headers
        self.headers = {k.decode(): v.decode() for k, v in scope.get("headers", [])}
        # Parse query params
        qs = scope.get("query_string", b"").decode()
        self.query = {k: v[0] if len(v) == 1 else v for k, v in parse_qs(qs).items()}
        # Parse cookies
        raw = self.headers.get("cookie", "")
        self.cookies = dict(pair.strip().split("=", 1) for pair in raw.split(";") if "=" in pair)

    async def body(self):
        chunks = []
        while True:
            msg = await self._receive()
            chunks.append(msg.get("body", b""))
            if not msg.get("more_body"): break
        return b"".join(chunks)

    async def json(self):
        import json
        return json.loads(await self.body())
